// SincResize
// a DFT-based upsizer for avisynth
// 2007 CAFxX (http://cafxx.strayorange.com)

#include <windows.h>
#include <crtdbg.h>
#include "..\..\src\core\avisynth.h"
#include <ippcore.h>
#include <ippi.h>
#include <omp.h>

IppiSize ROI(int W, int H) { IppiSize S; S.height = H; S.width = W; return S; }

class SincResize : public GenericVideoFilter {
	IppiDCTFwdSpec_32f* DCT;
	IppiDCTInvSpec_32f* iDCT;
	Ipp8u*	IPPTBuffer;	int IPPTBufferAlign;
	Ipp8u*	IPPPad;		int IPPPadAlign;
	Ipp32f* IPPIn;		int IPPInAlign;
	Ipp32f* IPPOut;		int IPPOutAlign;
	Ipp32f* IPPBuffer;	int IPPBufferAlign;
	int Factor;
	int Window;

	void PlaneTransform(const BYTE* src, BYTE* dst, int srcpitch, int dstpitch, int srcwidth, int dstwidth, int srcheight, int dstheight);

public:
	SincResize(PClip, int, int, IScriptEnvironment*);
	~SincResize();
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

SincResize::SincResize(PClip _child, int _Factor, int _Window, IScriptEnvironment* env) : GenericVideoFilter(_child) {
	char UnknownError[] = "SincResize: Initialization error (Unknown error)";
	char MemoryError[]  = "SincResize: Initialization error (Memory allocation error)";
	char IPPError[]     = "SincResize: Initialization error (IPP error: %s)";
	IppStatus Errno;
	char* Error = UnknownError;

	// parameter validation
	if ((Factor = _Factor) < 2)
		env->ThrowError("SincResize: Parameter <Factor> should be greater than 1");
	if ((Window = (1 << _Window)) < 4)
		env->ThrowError("SincResize: Parameter <Window> should be greater than 1");
	if (vi.IsPlanar() == false)
		env->ThrowError("SincResize: Input <Clip> must be planar");
	if (vi.width % Window || vi.height % Window)
		env->ThrowError("SincResize: Input <Clip> dimensions should be mod 2^<Window> (%d)", Window);
	// init OpenMP
//	omp_set_num_threads(omp_get_num_procs());
	// init IPP
	Errno = ippStaticInit();
	// allocate memory needed by IPP and set up transforms
	Error = MemoryError;
	if ( IPPBuffer = ippiMalloc_32f_C1(Window*Factor, Window*Factor, &IPPBufferAlign) ) {
		if ( IPPIn = ippiMalloc_32f_C1(vi.width+Window, vi.height+Window, &IPPInAlign) ) {
			if ( IPPOut = ippiMalloc_32f_C1(vi.width*Factor, vi.height*Factor, &IPPOutAlign) ) {
				if ( IPPPad = ippiMalloc_8u_C1(vi.width+Window, vi.height+Window, &IPPPadAlign) ) {
					Error = IPPError;
					if ( !(Errno = ippiDCTFwdInitAlloc_32f(&DCT, ROI(Window,Window), ippAlgHintAccurate)) ) {
						if ( !(Errno = ippiDCTInvInitAlloc_32f(&iDCT, ROI(Window*Factor,Window*Factor), ippAlgHintAccurate)) ) {
							int DCTBufferSize, iDCTBufferSize;
							if ( !(Errno = ippiDCTFwdGetBufSize_32f(DCT, &DCTBufferSize)) ) {
								if ( !(Errno = ippiDCTInvGetBufSize_32f(iDCT, &iDCTBufferSize)) ) {
									Error = MemoryError;
									if ( IPPTBuffer = ippiMalloc_8u_C1(max(DCTBufferSize, iDCTBufferSize), 1, &IPPTBufferAlign) ) {
										vi.width *= Factor;
										vi.height *= Factor;
										return;
									}
									// nothing to unwind here
								}
								// nothing to unwind here
							}
							ippiDCTInvFree_32f(iDCT);
						}
						ippiDCTFwdFree_32f(DCT);
					}
					ippiFree(IPPPad);
				} 
				ippiFree(IPPOut);
			}
			ippiFree(IPPIn);
		}
		ippiFree(IPPBuffer);
	}
	if (Error == IPPError)
		env->ThrowError(Error, ippGetStatusString(Errno));
	else
		env->ThrowError(Error);
}

SincResize::~SincResize() {
    ippiDCTFwdFree_32f(DCT);
    ippiDCTInvFree_32f(iDCT);
	ippiFree(IPPBuffer);
	ippiFree(IPPTBuffer);
	ippiFree(IPPOut);
	ippiFree(IPPIn);
	ippiFree(IPPPad);
}

void SincResize::PlaneTransform(const BYTE* src, BYTE* dst, int srcpitch, int dstpitch, int srcwidth, int dstwidth, int srcheight, int dstheight) {
	int x, y;

	ippiCopyReplicateBorder_8u_C1R(src, srcpitch, ROI(srcwidth, srcheight), IPPPad, IPPPadAlign, ROI(srcwidth+Window, srcheight+Window), Window/2, Window/2);
	ippiConvert_8u32f_C1R(IPPPad, IPPPadAlign, IPPIn, IPPInAlign, ROI(srcwidth+Window, srcheight+Window));
	ippiMulC_32f_C1IR((float)Factor, IPPIn, IPPInAlign, ROI(srcwidth+Window, srcheight+Window));
//	#pragma omp parallel for schedule(static)
	for (y=0; y<srcheight; y+=Window/2) {
		for (x=0; x<srcwidth; x+=Window/2) {
			memset(IPPBuffer, 0, IPPBufferAlign*Window*Factor);
			ippiDCTFwd_32f_C1R(IPPIn+x+Window/4+(y+Window/4)*IPPInAlign/sizeof(Ipp32f), IPPInAlign, IPPBuffer, IPPBufferAlign, DCT, IPPTBuffer);
			ippiDCTInv_32f_C1R(IPPBuffer, IPPBufferAlign, IPPBuffer, IPPBufferAlign, iDCT, IPPTBuffer);
			ippiCopy_32f_C1R(IPPBuffer+Window*Factor/4*(1+IPPBufferAlign/sizeof(Ipp32f)), IPPBufferAlign, IPPOut+x*Factor+y*Factor*IPPOutAlign/sizeof(Ipp32f), IPPOutAlign, ROI(Window*Factor/2,Window*Factor/2));
		}
	}
	ippiConvert_32f8u_C1R(IPPOut, IPPOutAlign, dst, dstpitch, ROI(dstwidth, dstheight), ippRndNear);
}

PVideoFrame __stdcall SincResize::GetFrame(int n, IScriptEnvironment* env) {
	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);
	
	PlaneTransform(src->GetReadPtr(PLANAR_Y), dst->GetWritePtr(PLANAR_Y), src->GetPitch(PLANAR_Y), dst->GetPitch(PLANAR_Y), src->GetRowSize(PLANAR_Y), dst->GetRowSize(PLANAR_Y), src->GetHeight(PLANAR_Y), dst->GetHeight(PLANAR_Y));
	PlaneTransform(src->GetReadPtr(PLANAR_U), dst->GetWritePtr(PLANAR_U), src->GetPitch(PLANAR_U), dst->GetPitch(PLANAR_U), src->GetRowSize(PLANAR_U), dst->GetRowSize(PLANAR_U), src->GetHeight(PLANAR_U), dst->GetHeight(PLANAR_U));
	PlaneTransform(src->GetReadPtr(PLANAR_V), dst->GetWritePtr(PLANAR_V), src->GetPitch(PLANAR_V), dst->GetPitch(PLANAR_V), src->GetRowSize(PLANAR_V), dst->GetRowSize(PLANAR_V), src->GetHeight(PLANAR_V), dst->GetHeight(PLANAR_V));

	return dst;
}

AVSValue __cdecl Create_SincResize(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new SincResize(args[0].AsClip(), args[1].AsInt(2), args[2].AsInt(3), env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) { 
    env->AddFunction("SincResize", "c[factor]i[window]i", Create_SincResize, 0);
	return "SincResize";
}

